#include "precomp.h"
#include "Smith#.h"
#include "opp.h"

#define SKIP_WS()                           \
    while( (*m_lpszExpression == ' ') ||    \
           (*m_lpszExpression == '\r') ||   \
           (*m_lpszExpression == '\n') ||   \
           (*m_lpszExpression == '\t') )    \
        m_lpszExpression++;

#define ENUMERATE_LABELS(ROOT,VARIABLE) \
    SFPLabel* VARIABLE = ROOT;          \
    for( ; VARIABLE; VARIABLE = VARIABLE->m_lpNext )

CellPU CPU;

CellPU::CellPU()
{
    m_nCount = sizeof(m_Cells)/sizeof(CELL);
}


SFPLabel::SFPLabel( SFPLabel** lpRoot, char* lpszName, long lOffset )
    :   m_lOffset( lOffset )
{
    m_lpNext = *lpRoot;
    *lpRoot = this;
    m_lpszName = strdup(lpszName);
} 

SFPLabel::~SFPLabel()
{
    if( m_lpszName )
        free( m_lpszName );

    delete m_lpNext;
}

bool StraightFileParser::Read( char* lpszFileName )
{
    PString strFilename(0,"%s.opp",lpszFileName);
    char* pExt = strrchr(strFilename,'.');
    if(pExt)
    {
        *pExt = ',';

        OPP opp;
        if( !opp.ProcessFile(lpszFileName,strFilename) )
            return false;

        lpszFileName = strFilename;
    }
    FILE* fp = fopen(lpszFileName, "r" );
    if( fp )
    {
        fseek(fp,0,SEEK_END);
        long fs = ftell(fp);
        char* data = 0;
        if( fs > 0 )
        {
            data = (char*) calloc(sizeof(char),fs+32);
            if( data )
            {
                fseek(fp,0,SEEK_SET);
                fread(data,sizeof(char),fs+1,fp);
            }
        }
        fclose(fp);

        if( data )
        {
            bool bCommentMode = false;
            char* rp = data;
            for( ; rp < (data+fs) && *rp; rp++ )
            {
                if( bCommentMode )
                {
                    if( *rp == '\n' )
                        bCommentMode = false;
                    else
                        *rp = ' ';
                }
                else
                {
                    if( *rp == ';' )
                    {
                        bCommentMode = true;
                        *rp = ' ';
                    }
                }
            } 

            m_lWritePos = 0;
            m_lpLabels = 0;
            m_lpReferences = 0;
            m_lpEvals = 0;
            rp = data;
            while( rp < (data+fs) && *rp )
            {
                m_lpszExpression = rp;
                SKIP_WS();
                if( !*m_lpszExpression )
                    break;

                if( !PARSE_EXPRESSION() )
                {
                    free(data);
                    return false;
                }
                rp = m_lpszExpression;
            }
            free(data);
            return true;
        }
    }
    else printf( "ERROR, unable to read file %s\n", lpszFileName );
    
    return false;
}

bool StraightFileParser::NORMCODE(long* pl)
{
    static char* SupportedCodes[] = 
    { "=", "<>", "<=", ">=", "<", ">", 0};
    static int nCodeSize[] =
    { 1, 2, 2, 2, 1, 1 };
    static int nNormcodes[] =
    { NORMCODE_EQUAL, NORMCODE_NOT_EQUAL, NORMCODE_LESS_EQUAL, 
      NORMCODE_GREATER_EQUAL, NORMCODE_LESS, NORMCODE_GREATER };

    for( int i = 0; SupportedCodes[i]; i++ )
    {
        if( !strncmp(m_lpszExpression,SupportedCodes[i],nCodeSize[i]) )
        {
            *pl = nNormcodes[i];
            m_lpszExpression += nCodeSize[i];
            return true;
        }
    }
    return false;
}


bool StraightFileParser::PARSE_EXPRESSION() 
{
    SKIP_WS();
    LABEL();

    SKIP_WS();
    if( !*m_lpszExpression )
        return true;

    if( OPCODE() )
    {
        bool bIsCiteOp = CPU.m_Cells[m_lWritePos-1] == OPCODE_CITE;

        if( CPU.m_Cells[m_lWritePos-1] == OPCODE_STOP )
            return true;

        SKIP_WS();
        if( ( bIsCiteOp && ASCIIOP() ) || OPERAND() )
        {
            while(1)
            {
                SKIP_WS();
                if( *m_lpszExpression == ',' )
                {
                    m_lpszExpression++;
                    SKIP_WS();
                    if( !(( bIsCiteOp && ASCIIOP() ) || OPERAND() ) )
                        return false;
                }
                else break;
            }
            SKIP_WS();
            return true;
        }
    }
    return false;
}

void StraightFileParser::LABEL()
{
    char* p = m_lpszExpression;

    int n;

    if( CHAR(&n) || DECDIGIT(&n) )
    {
        while( CHAR(&n) || DECDIGIT(&n) );
        SKIP_WS();
        if( *m_lpszExpression == ':' )
        {
            *(m_lpszExpression++) = 0;
            SKIP_WS();
            new SFPLabel(&m_lpLabels,p,m_lWritePos);
            return;
        }
    }
    m_lpszExpression = p;
}

bool StraightFileParser::CHAR(int* pc)
{
    char c = *m_lpszExpression;
    if( ((c >= 'a') && (c <= 'z')) ||
        ((c >= 'A') && (c <= 'Z')) ||
        ( c == '_' ))
    {
        *pc = c;
        m_lpszExpression++;
        return true;
    }
    return false;
}

bool StraightFileParser::DECDIGIT(int* pc)
{
    char c = *m_lpszExpression;
    if( (c >= '0') && (c <= '9') )
    {
        *pc = c - '0';
        m_lpszExpression++;
        return true;
    }
    return false;
}

bool StraightFileParser::DECINT(long* pl)
{
    int i;
    if( DECDIGIT(&i) )
    {
        *pl = i;
        while( DECDIGIT(&i) )
        {
            *pl *= 10;
            *pl += i;
        }
        return true;
    }
    return false;
}

bool StraightFileParser::IDENTIFIER(char** ppszName)
{
    char* p = m_lpszExpression;
    *ppszName = 0;
    int n;

    if( CHAR(&n) )
    {
        while( CHAR(&n) || DECDIGIT(&n) );

        char c = *m_lpszExpression;
        *m_lpszExpression = 0;
        *ppszName = strdup(p);
        *m_lpszExpression = c;
        return true;
    }
    return false;
}

bool StraightFileParser::OPCODE()
{
    if( DECINT( &CPU.m_Cells[m_lWritePos] ) )
    {
        m_lWritePos++;
        return true;
    }

    if( EVAL() )
        return true;

    if( OPCODEID(&CPU.m_Cells[m_lWritePos]) || NORMCODE(&CPU.m_Cells[m_lWritePos]) )
    {
        m_lWritePos++;
        return true;
    }

    bool bIsIndirect = false;
    SKIP_WS();
    if( *m_lpszExpression == '^' )
    {
        bIsIndirect = true;
        m_lpszExpression++;
        SKIP_WS();
    }

    char* pszName;
    if( IDENTIFIER(&pszName) )
    {
        new SFPLabel(&m_lpReferences, pszName,bIsIndirect ? -m_lWritePos : m_lWritePos);
        free(pszName);
        CPU.m_Cells[m_lWritePos++] = 0xBAADBEEF;
        return true;
    }

    // missing: EVAL
    return false;
}

bool StraightFileParser::OPERAND()
{
    return OPCODE();
}

bool StraightFileParser::OPCODEID(long* pl)
{
    static char* Codes[] =       
    {
        "STOP", ".", "CITE", "\"", "COPY", ":=", "NORM", "@", "ADD", "+", "SUB","-","MUL", "*", "DIV","/",
        "INPUT", ">>", "OUTPUT", "<<", "NAND", "~&", 0 };

    static int nOpcodes[] =
    {
        OPCODE_STOP, OPCODE_STOP, OPCODE_CITE, OPCODE_CITE, OPCODE_COPY, OPCODE_COPY, OPCODE_NORM, OPCODE_NORM, 
        OPCODE_ADD, OPCODE_ADD, OPCODE_SUB, OPCODE_SUB, OPCODE_MUL, OPCODE_MUL,
        OPCODE_DIV, OPCODE_DIV, OPCODE_INPUT, OPCODE_INPUT, OPCODE_OUTPUT, OPCODE_OUTPUT,
        OPCODE_NAND, OPCODE_NAND 
    };

    for( int i = 0; Codes[i]; i++ )
    {
        if( !strnicmp(m_lpszExpression,Codes[i],strlen(Codes[i])) )
        {
            *pl = nOpcodes[i];
            m_lpszExpression += strlen(Codes[i]);

            return true;
        }
    }
    return false;
}

bool StraightFileParser::ASCIIOP()
{
    if( *m_lpszExpression == '"' )
    {
        int nStringCount = 0;
        m_lpszExpression++; 
        while( *m_lpszExpression && (*m_lpszExpression != '"') )
        {
            char c = *(m_lpszExpression++);
            CPU.m_Cells[m_lWritePos++] = c;
            CPU.m_Cells[m_lWritePos++] = OPCODE_CITE;
        }
        if( *m_lpszExpression )
            m_lpszExpression++;
        CPU.m_Cells[m_lWritePos++] = 0;
        return true;
    }
    else if( m_lpszExpression[0] == '\'' && m_lpszExpression[2] == '\'')
    {
        CPU.m_Cells[m_lWritePos++] = m_lpszExpression[1];
        m_lpszExpression += 3;
        return true;
    }
    return false;
}

bool StraightFileParser::EVAL()
{
    bool bIsEval = false;
    if( !strnicmp(m_lpszExpression,"EVAL",4) )
        m_lpszExpression += 4, bIsEval = true;

    if( !strnicmp(m_lpszExpression,"?",1) )
        m_lpszExpression += 1, bIsEval = true;

    if( bIsEval )
    {
        SKIP_WS();
        if( *m_lpszExpression == '(' )
        {
            char* p = ++m_lpszExpression;
            SKIP_WS();
            if( EXP0(0) )
            {
                SKIP_WS();
                if( *m_lpszExpression == ')' )
                {
                    *m_lpszExpression = 0;
                    new SFPLabel(&m_lpEvals,p,m_lWritePos);
                    CPU.m_Cells[m_lWritePos++] = 0xFFFFFFFF;
                    m_lpszExpression++;
                    SKIP_WS();
                    return true;
                }
            }
        }

    }
    return false;
}

bool StraightFileParser::EXP0(long* pl)
{
    bool bNegative = false;
    if( *m_lpszExpression == '+' ) 
        m_lpszExpression++;
    else if( *m_lpszExpression == '-')
        m_lpszExpression++, bNegative = true;

    SKIP_WS();

    if( EXP1(pl) )
    {
    
        if( pl && bNegative )
            *pl = -*pl;
        
        while(1)
        {
            SKIP_WS();
            if( *m_lpszExpression == '+' ) 
                m_lpszExpression++, bNegative = false;
            else if( *m_lpszExpression == '-')
                m_lpszExpression++, bNegative = true;
            else return true;
            SKIP_WS();
            long temp;
            if( !EXP1(pl ? &temp : 0) )
                return false;

            if( pl )
                *pl = bNegative ? (*pl - temp) : (*pl + temp);
        }
    }
    return false;
}

bool StraightFileParser::EXP1(long* pl)
{
    SKIP_WS();
    if( EXP2(pl) )
    {
        bool bMultiply;
        while(1)
        {
            SKIP_WS();
            if( *m_lpszExpression == '/' ) 
                m_lpszExpression++, bMultiply = false;
            else if( *m_lpszExpression == '*')
                m_lpszExpression++, bMultiply = true;
            else return true;
            SKIP_WS();
        
            long temp;
            if( !EXP2(pl ? &temp : 0) )
                return false;
            
            if( pl && !bMultiply && !temp )
                return false; // DIVISION BY ZERO

            if( pl )
                *pl = bMultiply ? (*pl * temp) : (*pl / temp);
        }
        return true;
    }
    return false;
}

bool StraightFileParser::EXP2(long* pl)
{
    if( EXP3(pl) )
        return true;
    
    SKIP_WS();
    if( *m_lpszExpression == '(' )
    {
        m_lpszExpression++;
        SKIP_WS();
        if( EXP0(pl) )
        {
            SKIP_WS();
            if( *m_lpszExpression == ')' )
            {
                m_lpszExpression++;
                SKIP_WS();
                return true;
            }
        }
    }
    return false;
}

bool StraightFileParser::EXP3(long* pl)
{
    long unused, *ps;
    ps = pl ? pl : &unused;

    if( DECINT(ps) || NORMCODE(ps) || OPCODEID(ps) )
        return true;

    char* pszName;
    if( IDENTIFIER(&pszName) )
    {
        if( !pl )
        {
            free(pszName);
            return true;
        }

        ENUMERATE_LABELS( m_lpLabels, pLabel )
        {
            if( stricmp( pLabel->m_lpszName, pszName ) == 0 )
            {
                free(pszName);
                if( pl ) *pl = pLabel->m_lOffset;
                return true;
            }
        }
    }
    return false;
}

bool StraightFileParser::ResolveReferences()
{
    ENUMERATE_LABELS( m_lpReferences, pReference )
    {
        ENUMERATE_LABELS( m_lpLabels, pLabel )
        {
            if( stricmp( pLabel->m_lpszName, pReference->m_lpszName ) == 0 )
            {
                if( pReference->m_lOffset >= 0 )
                {
                    CPU.m_Cells[pReference->m_lOffset] = pLabel->m_lOffset + 1;
                }
                else
                {
                    CPU.m_Cells[-pReference->m_lOffset] = -(pLabel->m_lOffset + 1);
                }
                break;
            }
        }
        if( !pLabel )
        {
            printf( "*** ERROR, unable to resolve label %s\n", pReference->m_lpszName );
            return false;
        }
    }

    ENUMERATE_LABELS( m_lpEvals, pEval )
    {
        m_lpszExpression = pEval->m_lpszName;
        long result;
        if( EXP0(&result) )
        {
            CPU.m_Cells[pEval->m_lOffset] = result;
        }
        else
        {
            printf( "*** ERROR, unable to resolve expression %s\n", pEval->m_lpszName );
            return false;
        }
    }
    return true;
}

bool CellPU::Run()
{
    m_nIP = 0;
    while( m_Cells[m_nIP] != OPCODE_STOP )
    {
        if( !SingleStep() )
            return false;
    }
    printf( "\r\n" );
    return true;
}


#define GetNextOpcode(NAME)             \
        long NAME = m_Cells[m_nIP++];   \
        if( NAME < 0 )                  \
            NAME = m_Cells[-NAME];

#define GetNextOpcodeU(NAME)             \
        long NAME = m_Cells[pseudoIP++];   \
        if( NAME < 0 )                  \
            NAME = m_Cells[-NAME];

bool CellPU::Unassemble(long from, long to)
{
    long pseudoIP = from;
    long maxIP = from + to;

    while(1) 
    {
        if( pseudoIP >= maxIP )
            break;

        if( pseudoIP >= m_nCount )
        {
            printf( "*** ERROR, RAN OUT OF MEMORY\n" );
            return false;
        }

        printf( "%5d: ", pseudoIP );

        switch(m_Cells[pseudoIP++])
        {
        case OPCODE_STOP:
            printf( "STOP\n" );
            break;

        case OPCODE_CITE:
            printf( "CITE %d\n", m_Cells[pseudoIP++] );
            break;

        case OPCODE_COPY: // three parameters : from, to, skip
            {
                GetNextOpcodeU(from);
                GetNextOpcodeU(to);
                long size = m_Cells[m_Cells[m_nIP++]];

                printf( "COPY %d,%d,%d\n", from, to, size );
            }
            break;

        case OPCODE_NORM:
            {
                GetNextOpcodeU(code);
                GetNextOpcodeU(valix);
                long value = m_Cells[valix];
                long factor = m_Cells[m_Cells[pseudoIP++]];

                printf( "NORM %d,%d,%d\n", code,value,factor);
            }
            break;
        case OPCODE_ADD:
            {
                GetNextOpcodeU(fromix);
                GetNextOpcodeU(toix);

                printf( "ADD %d,%d\n", fromix, toix );
            }
            break;
        case OPCODE_SUB:
            {
                GetNextOpcodeU(fromix);
                GetNextOpcodeU(toix);

                printf( "SUB %d,%d\n", fromix, toix );
            }
            break;
        case OPCODE_MUL:
            {
                GetNextOpcodeU(fromix);
                GetNextOpcodeU(toix);

                printf( "MUL %d,%d\n", fromix, toix );
            }
            break;
        case OPCODE_DIV:
            {
                GetNextOpcodeU(fromix);
                GetNextOpcodeU(toix);

                printf( "DIV %d,%d\n", fromix, toix );
            }
            break;
        case OPCODE_OUTPUT:
            {
                GetNextOpcodeU(outx);
                printf( "OUTPUT %d\n", m_Cells[outx] );
            }
            break;
        case OPCODE_INPUT:
            {
                GetNextOpcodeU(outx);
                printf( "INPUT %d\n", m_Cells[outx] );
            }
            break;
        case OPCODE_NAND:
            {
                GetNextOpcodeU(fromix);
                GetNextOpcodeU(toix);

                printf( "NAND %d,%d\n", fromix, toix );
            }
            break;

        default:
            pseudoIP--;
            printf( "*** ERROR, UNKNOWN INSTRUCTION %d AT IP %d\n", m_Cells[pseudoIP], pseudoIP );
            return false;
        }
    }
    return true;

}

bool CellPU::SingleStep()
{
    if( m_nIP >= m_nCount )
    {
        printf( "*** ERROR, RAN OUT OF MEMORY\n" );
        return false;
    }
    if( m_bDebugMode )
        printf( "%3d: ", m_nIP );

    switch(m_Cells[m_nIP++])
    {
    case OPCODE_STOP:
        printf( "STOP\n" );
        m_nIP--;
        return true;

    case OPCODE_CITE:
        if( m_bDebugMode )
            printf( "cite %d\n", m_Cells[m_nIP] );
        m_nIP++;
        break;

    case OPCODE_COPY: // three parameters : from, to, skip
        {
            GetNextOpcode(from);
            GetNextOpcode(to);
            long size = m_Cells[m_Cells[m_nIP++]];

            if( m_bDebugMode )
                printf( "Copy %d cells from %d to %d\n", size, from, to );
            
            for( long index = 0; index < size; index++ )
                m_Cells[to++] = m_Cells[from++];
        }
        break;

    case OPCODE_NORM:
        {
            GetNextOpcode(code);
            GetNextOpcode(valix);
            long value = m_Cells[valix];
            long factor = m_Cells[m_Cells[m_nIP++]];

            if( m_bDebugMode )
                printf( "Evaluate %d using code %d, factor %d\n", value, code, factor );

            switch(code)
            {
            case NORMCODE_EQUAL:
                m_Cells[valix] = (value == 0) ? factor : 0;
                break;
            case NORMCODE_NOT_EQUAL:
                m_Cells[valix] = (value != 0) ? factor : 0;
                break;
            case NORMCODE_LESS:
                m_Cells[valix] = (value < 0) ? factor : 0;
                break;
            case NORMCODE_GREATER:
                m_Cells[valix] = (value > 0) ? factor : 0;
                break;
            case NORMCODE_LESS_EQUAL:
                m_Cells[valix] = (value <= 0) ? factor : 0;
                break;
            case NORMCODE_GREATER_EQUAL:
                m_Cells[valix] = (value >= 0) ? factor : 0;
                break;
            default:
                printf( "*** ERROR, UNKNOWN NORMCODE %d AT IP %d", code, m_nIP-2 );
                return false;
            }
            if( m_bDebugMode )
                printf( "Evaluates to %d\n", m_Cells[valix] );

        }
        break;
    case OPCODE_ADD:
        {
            GetNextOpcode(fromix);
            GetNextOpcode(toix);

            if( m_bDebugMode )
                printf( "Add cell %d to cell %d\n", fromix, toix);
        
            m_Cells[toix] += m_Cells[fromix];
        }
        break;
    case OPCODE_SUB:
        {
            GetNextOpcode(fromix);
            GetNextOpcode(toix);

            if( m_bDebugMode )
                printf( "Subtract cell %d from cell %d\n", fromix, toix);
        
            m_Cells[toix] -= m_Cells[fromix];
        }
        break;
    case OPCODE_MUL:
        {
            GetNextOpcode(fromix);
            GetNextOpcode(toix);

            if( m_bDebugMode )
                printf( "Multiply cell %d with cell %d\n", toix, fromix);
        
            m_Cells[toix] *= m_Cells[fromix];
        }
        break;
    case OPCODE_DIV:
        {
            GetNextOpcode(fromix);
            GetNextOpcode(toix);

            if( m_bDebugMode )
                printf( "Divide cell %d by cell %d\n", toix, fromix );

            if( m_Cells[fromix] )
                m_Cells[toix] /= m_Cells[fromix];
            else
            {
                printf( "*** ERROR, DIVISION BY ZERO AT IP %d\n", m_nIP-1 );
                return false;
            }
        }
        break;
    case OPCODE_OUTPUT:
        {
            if( m_bDebugMode )
                printf( "Output character from cell %d\n", m_Cells[m_nIP] );

            GetNextOpcode(outx);
            printf("%c",(char) m_Cells[outx]);
            if( m_bDebugMode )
                printf( "\n" );
        }
        break;
    case OPCODE_INPUT:
        {
            printf( "Input character to cell %d\n", m_Cells[m_nIP] );
            m_nIP += 1;
        }
        break;
    case OPCODE_NAND:
        {
            GetNextOpcode(fromix);
            GetNextOpcode(toix);

            m_Cells[toix] = (~(m_Cells[fromix])) & (~(m_Cells[toix]));
        }
        break;

    default:
        m_nIP--;
        printf( "*** ERROR, UNKNOWN INSTRUCTION %d AT IP %d\n", m_Cells[m_nIP], m_nIP );
        return false;
    }
    return true;
}

int main( int argc, char* argv[] )
{
    printf( "Welcome to SMITH#, Version 1.1\n"
            "Type ? for help\n" );

    char buffer[256];
    CPU.m_bDebugMode = false;
    CPU.m_nIP = 0;
    while(1)
    {
        printf( ":) " );
        gets(buffer);
        
        if( !buffer[0] )
            continue;

        switch(toupper(buffer[0]))
        {
        case 'X':
        case 'Q':
            return 0;
        case 'Z':
            {
                CPU.m_nIP = 0;
                for( int i = 0; i < CPU.m_nCount; i++ )
                    CPU.m_Cells[i] = 0;
            }
            break;
        case 'C':
            {
                StraightFileParser sfp;
                if( sfp.Read(buffer+1) )
                    if( sfp.ResolveReferences() )
                        printf("Code uses %d cells\n", sfp.m_lWritePos );
            }
            break;
        case 'G':   
            CPU.m_nIP = 0;
            CPU.Run();
            break;
        case 'S':   
            CPU.SingleStep();
            break;
        case 'M':
            {
                char* p = strchr(buffer+1,':');
                if( p )
                {
                    *p++ = 0;
                    long from = atol(buffer+1);
                    long size = atol(p);
                    long to = from + size;

                    for( long index = 0; index < size; index++ )
                    {
                        printf( "%ld, ", CPU.m_Cells[from+index] );
                    }
                    printf( "\n" );
                }
            }
            break;
        case 'D':
            CPU.m_bDebugMode = CPU.m_bDebugMode ? false : true;
            printf( "Debug mode %s\n", CPU.m_bDebugMode ? "on" : "off" );
            break;
        case 'U':
            {
                char* p = strchr(buffer+1,':');
                if( p )
                {
                    *p++ = 0;
                    long from = atol(buffer+1);
                    long size = atol(p);
                    CPU.Unassemble(from,size);
                }
            }
            break;
        case '?':
            printf( "X,Q ............... Quit / eXit\n"
                    "Z ................. Zap everything (resets all data)\n" 
                    "C<Filename> ....... read and Compile sourcecode\n"
                    "G ................. Go (execute sourcecode)\n"
                    "D ................. Toggle debug mode (defaults to OFF)\n"
                    "M<from>:<size>..... Dump cells\n"
                    "U<from>:<size>..... Unassemble cells\n"
                    "S ................. Single step\n" );
        }
    }
    return 0;
} // main()



